function [p,y1,y2,han1]   =   QP(y,options)  

% -------------------------------------------------------------------------
% Qualitative Path Estimation Toolbox - QP.m
% -------------------------------------------------------------------------
%
%   This function computes the point-wise Qualitative state Probabilities
%   (QP).
%
%   I/O: [p,y1,y2,han1]   =   QP(y,options) 
%
%   inputs:
%       y:          data vector
%       options:    structure specifying options for the algorithm   
%
%   outputs:
%       p:          Point-wise probabilities
%       y1:         Filtered first derivative
%       y2:         Filtered second derivative
%       han1:       handles for figure axes
%
% -------------------------------------------------------------------------
% Last modification: Kris Villez, 2015-01-11
% -------------------------------------------------------------------------

% -------------------------------------------------------------------------
% Copyright 2013-2015 Kris Villez
%
% This file is part of the QPE Toolbox for Matlab/Octave. 
% 
% The QPE Toolbox is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by the
% Free Software Foundation, either version 3 of the License, or (at your
% option) any later version.
% 
% The QPE Toolbox is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
% Public License for more details.
% 
% You should have received a copy of the GNU General Public License along
% with the QPE Toolbox. If not, see <http://www.gnu.org/licenses/>.  
% -------------------------------------------------------------------------

% =========================================================================
% 1. SETUP 

% -------------------------------------------------------------------------
% 1.1 options for debugging / display
graph       =   1   ;
verbose     =   1   ;

% -------------------------------------------------------------------------
% 1.2 read in algorithm options
n_support   =   options.n_support   ;
uncfac 		=   options.uncfac      ;

% -------------------------------------------------------------------------
% 1.3 read in data
y 			= 	y(:) ;
ny      	=   length(y)       ;

% -------------------------------------------------------------------------
% 1.4 setup kernel regression filter
n_width =       n_support+2-mod(n_support,2) 	; % +1 for odd support window, +2 for even support window
lambda 	=       n_width/2 						; % lambda parameter: n_support gives the window of active data points

w       =       lambda-1                            ;
tk      =   [   -w:w  ]'                            ;
tt      =   [   -w:w  ]'/lambda                     ;

D       =       (1-abs(tt).^3).^3 .* (abs(tt)<=1)	;   % Tri-Cube:
%D       =       (1-abs(tt).^2)                      ;   % Quadratic

tt  	=       tt(find(D~=0))                      ;
D   	=       D(find(D~=0))                       ;
B   	=   [	tk.^0	tk.^1	tk.^2/2 tk.^3/6	tk.^4/24]           ;
order 	= 		3 				;
B 		= 		B(:,1:order) 	;

H0      =       (B'*diag(D)*B)^(-1)*B'*diag(D)      ;

Sy0     =       eye(length(tt))             		;
Sb0     =       H0*Sy0*H0'                          ;
sb0     =       diag(Sb0).^(1/2)                    ;

% -------------------------------------------------------------------------
% 1.5 scaling parameters for time and data, hard-coded for demo example
timefac =   (ny-1)/400	;
scale   =   0.0316      ;
      

% -------------------------------------------------------------------------
% 1.6 Initialize results
Y   =   nan(ny,order)	;
Yn  =   nan(ny,order)	;
SB  =   nan(ny,order)	;

% -------------------------------------------------------------------------
% 1.7 Initialize graph for kernel regression
graph =1;
if graph
    figure, % FigPrep
    subplot(3,1,1), 
        % AxisPrep, 
        hold on
        ty = (1:length(y))/timefac ;
        plot(ty,scale*y,'k.'), 
        ylabel({'y','[g/l]'}) ;
        set(gca,'Xticklabel',[],'Xlim',[ty(1)-1 ty(end)+1])
      
        subplot(3,1,2), 
        % AxisPrep, 
        hold on, 
        ylabel({'dy/dt','[(g/l) / min.]'}) ;
        set(gca,'Xticklabel',[],'Xlim',[ty(1)-1 ty(end)+1])
    subplot(3,1,3), 
        % AxisPrep, 
        hold on, 
        ylabel({'d^2y/dt^2','[(g/l) / (min.)^2]'}) ;
        set(gca,'Xlim',[ty(1)-1 ty(end)+1])
        xlabel('Time [h]')
end
% =========================================================================

% =========================================================================
% 2. ANALYSIS

if verbose, disp(['			Kernel regression ']), end

% -------------------------------------------------------------------------
% 2.1 Iterative loop
for k=1:ny
    
    if mod(k,100)==1 || k==ny
        if verbose, disp(['				data point ' num2str(k)]), end
    end
	
    kk      =   k  + (-w:w)			;
    inc     =   and(kk>=1,kk<=ny)	;
	
    if all(inc)
        % use pre-defined filter
        b 	=	H0*y(kk)    			;
		sb	= 	sb0 					;
    else
        % recompute filter for windows outside of data range
        inc	=	find(inc)   			;
        kk 	= 	kk(inc)    				;
        Bi	= 	B(inc,:)    			;
        Di	= 	diag(D(inc))			;
        H 	= 	(Bi'*Di*Bi)^(-1)*Bi'*Di	;

		Sy 	=	eye(length(inc))  		;
        Sb 	=	H*Sy*H'     			;
        sb	= 	diag(Sb).^(1/2)     	;

        b	= 	H*y(kk)    				;
    end
	
    yhat    = 	b                   ;
	yn      = 	diag(1./sb)*yhat    ;
	
	
    Y(k,:) 	= 	yhat(:)'	;
    Yn(k,:) =   yn(:)'		;
    
    SB(k,:)  = sb ;

end

y0 	= 	Y(:,1) ;
y1 =   	Y(:,2) ;
y2 =   	Y(:,3) ;

% -------------------------------------------------------------------------
% 2.2 Graph kernel regression
if graph
    for sub=1:3
        han1(sub) = subplot(3,1,sub);
        if sub==1
            plot(ty,(timefac*60)^(sub-1)*scale*Y(:,sub),'-','LineWidth',2,'Color',0.25*ones(1,3)),
			legend('y')
			%legend({'$\mathbf{y}$','$\mathbf{\hat{f}}$'},'Interpreter','Latex','fontsize',14)
        else
            plot(ty,(timefac*60)^(sub-1)*scale*Y(:,sub),'k-','LineWidth',2),
            plot(ty,(timefac*60)^(sub-1)*scale*(Y(:,sub)+3*SB(:,sub)),'k--','LineWidth',2),
            plot(ty,(timefac*60)^(sub-1)*scale*(Y(:,sub)-3*SB(:,sub)),'k--','LineWidth',2),
            if sub==2
				legend('f^{(1)}')
                %legend({'$\mathbf{\hat{f}}^{(1)}$','$\mathbf{\hat{f}}^{(1)} \pm \sigma_{f^{(1)}}$'},'Interpreter','Latex','fontsize',14)
            else
				legend('f^{(2)}')
                %legend({'$\mathbf{\hat{f}}^{(2)}$','$\mathbf{\hat{f}}^{(2)} +/- \sigma_{f^{(2)}}$'},'Interpreter','Latex','fontsize',14)
            end
        end
        set(gca,'Xlim',[-1 401])
        if sub>=2
            %PlotHorizontal(0,'k:')
            ylims = get(gca,'Ylim') ;
            plot([0 0],[ylims],'k:')
        end
    end
end

% -------------------------------------------------------------------------
% 2.3 Estimation standard deviation and corresponding scaling

if ~isnan(uncfac) % if nan, nothing is changed
	std_e 	= 	std([y-y0]) 			;
	Yn 		= 	Yn/(std_e*uncfac) 		;
end

% -------------------------------------------------------------------------
% 2.4 Probabilities for the primitives

% y0n =   Yn(:,1) ;
y1n =   Yn(:,2) ;
y2n =   Yn(:,3) ;

alpha 	=	1e-15 ;

p1_0    =   normpdf(y1n)    .* 0/normpdf(0) ;
p1_p    =   normcdf(y1n)                    ;
p1_n    =   normcdf(-y1n)                   ;
p1_sum 	=  	p1_p+p1_n 						;
p1_p 	=	(p1_p+alpha)./(p1_sum+2*alpha) 	;
p1_n 	=	(p1_n+alpha)./(p1_sum+2*alpha) 	;

p2_0    =   normpdf(y2n)    .* 0/normpdf(0) ;
p2_p    =   normcdf(y2n)                    ;
p2_n    =   normcdf(-y2n)                   ;
p2_sum 	=  	p2_p+p2_n 						;
p2_p 	=	(p2_p+alpha)./(p2_sum+2*alpha) 	;
p2_n 	=	(p2_n+alpha)./(p2_sum+2*alpha) 	;

p       =   [   p1_n.*p2_n  p1_n.*p2_0  p1_n.*p2_p  p1_0  p1_p.*p2_n  p1_p.*p2_0  p1_p.*p2_p ] ;
p       =   p./(sum(p,2)*ones(1,7)) ; % normalization

% ------------------------------------------
% this to avoid zero probabilities
% always do this at beginning so that repetitive calls to this function
% with different sequences have the same probabilities to work with
[nk,np] =   size(p)                             ;
alpha   =   1e-3                                ;
p       =   (1-alpha)*p + alpha*(ones(nk,np))   ;
% ------------------------------------------

% -------------------------------------------------------------------------
% 2.5 Graph probabilities for primitives
graph =0 ;
if graph
    figure, 
    FigPrep
    AxisPrep
	hold on,
    size(p)
    size(tt)
    plot(ty,p(:,[3]),':','color',.5*ones(1,3),'linewidth',3)
    plot(ty,p(:,[7]),'-','color',0*ones(1,3),'linewidth',3)
    plot(ty,p(:,[5]),'--','color',0*ones(1,3),'linewidth',3)
	plot(ty,p(:,[1]),'-','color',.5*ones(1,3),'linewidth',2)
    xlabel(['Time [h]'])
    ylabel(['L(Pr|y)'])
    legend({'A','B','C','D'})
    set(gca,'Xlim',[ty(1)-1 ty(end)+1],'Ylim',[-.01 1.01])
end

