function [LogLik,trans,y1n,y2n,ComputTime] = QPE(y,t,kernelwidths,uncertaintyfactors,strings)

% -------------------------------------------------------------------------
% Qualitative Path Estimation Toolbox - QPE.m
% -------------------------------------------------------------------------
%
%   This function computes the Qualitative Representation (QR) for each of
%   the given Qualitative Sequences.
%
%   I/O: [LogLik,trans,y1n,y2n,ComputTime] =
%               QPE(y,t,kernelwidths,uncertaintyfactors,strings,constraint) 
%
%   inputs:
%       y:                  Data vector
%       t:                  [optional] time vector
%       kernelwidths:       Number of supporting data points
%       uncertaintyfactors: Factor by which the standard deviation estimate
%                           is multiplied. When using NaN, it is assumed
%                           the standard deviation is known to be one (1).
%       strings:            The strings representing the Qualitative
%                           Sequences (QSs, cell array)
%       
%   outputs:
%       LogLik:             Log-Likelihood for each of the strings (QSs)
%       trans:              Identified transition points
%       y1n:                Normalized first derivative
%       y2n:                Normalized second derivative
%       ComputTime:         Time needed for computation in each step of the
%                           algorithm
%
% -------------------------------------------------------------------------
% Last modification: Kris Villez, 2015-01-11
% -------------------------------------------------------------------------

% -------------------------------------------------------------------------
% Copyright 2013-2015 Kris Villez
%
% This file is part of the QPE Toolbox for Matlab/Octave. 
% 
% The QPE Toolbox is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by the
% Free Software Foundation, either version 3 of the License, or (at your
% option) any later version.
% 
% The QPE Toolbox is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
% Public License for more details.
% 
% You should have received a copy of the GNU General Public License along
% with the QPE Toolbox. If not, see <http://www.gnu.org/licenses/>.  
% -------------------------------------------------------------------------

% =========================================================================
% 1. SETUP 

% -------------------------------------------------------------------------
% 1.1 Options for debugging / display
graph       =   1   ;
verbose     =   1   ;

% -------------------------------------------------------------------------
% 1.2 Read in data
y 			= 	y(:)        ;
ny      	=   length(y)   ;
if isempty(t)
    t       =   (1:ny)'     ;
end

% -------------------------------------------------------------------------
% 1.3 Read in correspondence character - primitive index

[Characters] = Alphabet([],3) ;
if ~iscell(strings) 
    StringsWasNotCell = true ;
    strings = {strings} ;
else
    StringsWasNotCell = false ;
end

% -------------------------------------------------------------------------
% 1.4 Convert strings to sequences of primitive indices
nSeq                =   length(strings)     ;
Seq                 =   cell(nSeq,1)        ;
for iSeq =1:nSeq
    [oo,Sequence]   =   ismember(strings{iSeq},Characters)  ;
    Seq{iSeq}       =   Sequence                            ;
end

% -------------------------------------------------------------------------
% 1.5 Initialize results
ComputTime  =   nan(length(kernelwidths),length(uncertaintyfactors),nSeq+1)    ;
n_kernel    =   length(kernelwidths)        ;
n_unc       =   length(uncertaintyfactors)  ;

% =========================================================================

% =========================================================================
% 2. ANALYSIS

% -------------------------------------------------------------------------
% Level A. For each kernel support window ...
for i_kernel=1:n_kernel
    
    if verbose, disp([' kernel width: ' num2str(i_kernel) ' of ' num2str(n_kernel)]), end
    
    % -------------------------------------------------------------------------
    % Level B. For each option for the noise standard deviation ...
    for i_unc=1:n_unc
        if verbose, disp(['  uncertainty factor: ' num2str(i_unc) ' of ' num2str(n_unc)]), end
        
        % -------------------------------------------------------------------------
        % 2.1 Local setup
        n_support 			= 	kernelwidths(i_kernel)		;
        uncfac 				= 	uncertaintyfactors(i_unc)	;
        
        options.n_support   =   n_support    	;
        options.uncfac 		=   1*uncfac       ;
        options.seq     	=   Seq             ;
        
        
        % -------------------------------------------------------------------------
        % 2.2 Compute point-wise probabilities
        
        tic
        
        [p,y1n,y2n,han1]   	=   QP(y,options)               ;
        p               =   [p  nan(ny,7) ]                 ;
        fac             =   2                               ;
        p(:,8)          =   sum(p(:,1:3),2)     /fac        ;
        p(:,9)          =   sum(p(:,4),2)                   ;
        p(:,10)         =   sum(p(:,5:7),2)     /fac        ;
        p(:,11)         =   sum(p(:,[1 5]),2)   /fac        ;
        p(:,12)         =   sum(p(:,[2 4 6]),2) /fac        ;
        p(:,13)         =   sum(p(:,[3 7]),2)   /fac        ;
        p(:,14)         =   sum(p(:,1:7),2)     /fac/fac    ;
        
        time                            =   toc  	;
        ComputTime(i_kernel,i_unc,1)    =   time    ;
        
        % -------------------------------------------------------------------------
        % Level C. For each Qualitative Sequence ...
        
        for iSeq = 1:nSeq
            
            if verbose, disp(['			Viterbi - ' strings{iSeq}]), end
            
            % -------------------------------------------------------------------------
            % 2.3 Make graph
            if graph
                ny = length(y) ;
                timefac = (ny-1)/400 ;
                ty = (t-1)/timefac ;
                timefac = (ny-1)/400 ;
                figure,
                % FigPrep
                han2(1) = subplot(2,1,1) ;
                % AxisPrep
                hold on,
                plot(ty,p(:,[3]),':','color',.3*ones(1,3),'linewidth',2.5)
                plot(ty,p(:,[7]),'-','color',0*ones(1,3),'linewidth',2.5)
                plot(ty,p(:,[5]),'--','color',0*ones(1,3),'linewidth',2.5)
                plot(ty,p(:,[1]),'-','color',.6*ones(1,3),'linewidth',2.5)
                plot(ty,p(:,[3]),':','color',.3*ones(1,3),'linewidth',2.5)
				ylabel(['Lambda'],'Interpreter','Latex','Fontsize',16)
                %ylabel(['${\Lambda}(\kappa|\mathbf{y})$'],'Interpreter','Latex','Fontsize',16)
                legend({'A','B','C','D'})
                set(gca,'Xlim',[ty(1)-1 ty(end)+1],'Ylim',[-.01 1.01],'Xticklabel',[])
                %             plot(p(:,[2:2:7]),':')
            end
            
            % -------------------------------------------------------------------------
            % 2.4 Viterbi path estimation
            
            tic
            
            p_seq 		= 	p(:,Seq{iSeq})                                  ;
            [L,path]	=   Viterbi(p_seq,1:length(Seq{iSeq}))              ;
            
            index 		=	find(and(diff(path)~=0,~isnan(diff(path))))     ;
            solution	=	(t(index-1)+t(index))/2                         ;
            
            time     	=   toc     ;
            
            ComputTime(i_kernel,i_unc,1+iSeq)	=   time    			;
            LogLik(i_kernel,i_unc,iSeq)       	=   L       			;
            trans{i_kernel,i_unc,iSeq}         	=   solution(:) 		;
            
            % -------------------------------------------------------------------------
            % 2.5 Add on to graph
            
            if graph
                han2(2) = subplot(2,1,2) ;
                % AxisPrep
                tplot    = ([ (t(:)-1/2) (t(:)+1/2) ]-1)'/timefac ;
                tplot   =   tplot(:) ;
                %                 tplot   =   min(max(tplot,0),(t(end)+1)/timefac) ;
                pathplot = [ path(:)' ; path(:)' ];
                pathplot =  pathplot(:) ;
                
                hold on
                stairs(tplot-1/timefac,pathplot,'k-','linewidth',3)
                set(gca,'Xlim',[ty(1)-1 ty(end)+1],'Ylim',[0.9 4.6])
                xlabel(['Time [h]'])
                %PlotVertical((solution-1)/timefac,'r--','linewidth',2)
                ylims = get(gca,'Ylim') ;
                plot((solution-1)/timefac*ones(1,2),[ylims],'r--','linewidth',2)
				ylabel(['s_{path}'])
                %ylabel(['$s_{path}$'],'Interpreter','Latex','Fontsize',18)
                segments = [0; solution; ny-1 ];
                for j=1:(length(segments)-1)
                    text((mean(segments(j+(0:1)))-1)/timefac,4.3,strings{iSeq}(j),'HorizontalAlignment','center','FontSize',16)
                end
                
                
                for sub=1:3
                    axes(han1(sub) )
                    %PlotVertical((solution-1)/timefac,'r--','linewidth',2)
                    ylims = get(gca,'Ylim') ;
                    plot((solution-1)/timefac*ones(1,2),[ylims],'r--','linewidth',2)
                    
                end
                
                for sub=1:2
                    axes(han2(sub) )
                    %PlotVertical((solution-1)/timefac,'r--','linewidth',2)
                    ylims = get(gca,'Ylim') ;
                    plot((solution-1)/timefac*ones(1,2),[ylims],'r--','linewidth',2)
                    
                end
            end
            
            
        end
    end
end

if StringsWasNotCell
    trans = trans{1} ;
end
