function [Lfin,path]   =   Viterbi(p,Seq)  

% -------------------------------------------------------------------------
% Qualitative Path Estimation Toolbox - Viterbi.m
% -------------------------------------------------------------------------
%
%   This function computes the Viterbi path
%
%   I/O: [Lfin,path]	=   Viterbi(p,Seq,constraint) 
%
%   inputs:
%       p:                  Point-wise probabilities
%       Seq:                Sequence of states
%       
%   outputs:
%       Lfin:               Log-Likelihood for each of the sequences
%       path:               Identified state path
%
% -------------------------------------------------------------------------
% Last modification: Kris Villez, 2015-01-11
% -------------------------------------------------------------------------

% -------------------------------------------------------------------------
% Copyright 2013-2015 Kris Villez
%
% This file is part of the QPE Toolbox for Matlab/Octave. 
% 
% The QPE Toolbox is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by the
% Free Software Foundation, either version 3 of the License, or (at your
% option) any later version.
% 
% The QPE Toolbox is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
% Public License for more details.
% 
% You should have received a copy of the GNU General Public License along
% with the QPE Toolbox. If not, see <http://www.gnu.org/licenses/>.  
% -------------------------------------------------------------------------

% =========================================================================
% 1. SETUP 

% -------------------------------------------------------------------------
% 1.1 Options for debugging / display
verbose = 0 ;

% -------------------------------------------------------------------------
% 1.2 Setup transition likelihood matrix

[nk,np]      =   size(p)   ;

% transition probability
pt      =       1               ;   % 1 makes it equally possible to stay or go

% select only relevant observation probabilities ( p(y_k|s_k) )
ns      =       length(Seq)     ;
p       =       p(:,Seq)        ;

% construct matrix ( p(s_k|s_k-1) )
Linear      =   true	;
if Linear % Use linear Markov chain
    Pt0  	= 	spdiags([  (1)*ones(ns,1)  pt*ones(ns,1)  ],0:1,ns,ns) ;
else
    Pt0   	=  	(1-pt)*eye(ns)+pt*ones(ns) ;
end

% convert to log-likelihoods
warning('off','MATLAB:log:logOfZero')
Ly      =       log10(p)              ;
Lt0      =       log10(Pt0)             ;
warning('on','MATLAB:log:logOfZero')

% -------------------------------------------------------------------------
% 1.3 Initialization

L0      =   [   0 -inf(1,ns-1) ]    ;
Ls      =       nan(1,ns)           ;
Path    =       nan(nk,ns)          ;
% =========================================================================

% =========================================================================
% 2. ANALYSIS

% -------------------------------------------------------------------------
% 2.1 Run Viterbi forward pass

if verbose,  disp(['			  forward: ']), end

for k=1:nk
    
    if mod(k,100)==1 || k==nk
        if verbose, disp(['				data point ' num2str(k)]), end
    end
    
    Lt = Lt0 ;
	
    for it=1: ns
        %disp(['Target state: ' num2str(it)])
        Lp  =   L0  ;
        for is=1:ns  
            if ~isinf(Lp(is))
                Lp(is)  =   Lp(is)+Lt(is,it)+Ly(k,it) ;
            end
        end
        [Lmax,Smax]     =   max(Lp)     ;
        Ls(it)          =   Lmax        ;
        if Lmax==-inf
            Path(k,it)  =   nan         ;
        else
            Path(k,it)  =   Smax        ;
        end
    end
    L0 = Ls ;
end

% -------------------------------------------------------------------------
% 2.2 Run Viterbi backward pass: re-trace best path

k           =   nk          ;
Path(k,:)   =   [ 1:ns ]    ;

if verbose, disp(['			  backward: ']), end
for k=nk-1:-1:1
    
    if mod(k,1000)==1 || k==nk
        if verbose, disp(['				data point ' num2str(k)]), end
    end
    inc             =   ~isnan(Path(k+1,:))     ;
    Path(k,inc)     =   Path(k,Path(k+1,inc))   ;
end

path    =   Path(:,end)     ;
Lfin    =   L0(end)         ;